/*********************************************************
 * Implementation  of the Papenberg at al. optical flow  *
 *********************************************************/

#include <iostream>
#include "PAP.h"

#ifndef MATLAB_MEX_FILE       //For Windows use: MEXFILE_FUNCTION
/*********************************
 * Check validity of a numerical *
 * value: throw a numerical      *
 * exception in case of problem. *
 *********************************/
void PAP::checkValue(double val, unsigned x, unsigned y, unsigned z)
{
	if (val != val) throw NaNException(x, y, z);
	else if (fabs(val) == HUGE_VAL) throw INDException(x, y, z);
}
#endif

/*******************************
 * 2x2 symmetric system solver 
 *******************************/
void PAP::solve2x2(double a11, double a12, double a22, double b1, double b2, double &x, double &y)
{
	double invdet = 1.0/(a11*a22-a12*a12);
	x = invdet*(a22*b1 - a12*b2);
	y = invdet*(a11*b2 - a12*b1);
}


/*************************************
 * implement a resize  la pnmrescale
 * it's separable: starts by resizing 
 * columns and then rows.            
 *************************************/
void PAP::resize(CImg<float> & source, CImg<float> & dest)
{
	unsigned xs = source.width, ys = source.height;
	unsigned xd = dest.width,   yd = dest.height;
	unsigned nz = source.depth, nv = source.dim; // we won't touch time and vectorial dimensions
	// create intermediary image, fill with 0's
	CImg<float> inter(xs, yd, nz, nv, 0.0);
	
	unsigned left = yd*ys, a = yd, b = ys;
	unsigned s = 0, d = 0, used = 0;

	// redim source columns from ys to yd pixels, save in sequence "inter"
	while (left > 0) {
		used = cimg::min(a,b); // fraction of the input we need
		left -= used; a -= used; b -= used;
		cimg_mapXZV(inter, x, z, v) inter(x, d, z, v) += source(x, s, z, v)*used;
		if (b == 0) {
			cimg_mapXZV(inter, x, z, v) inter(x, d, z, v) /= ys;
			b = ys; ++d;
		}
		if (a == 0) {
			a = yd; ++s;
		}
	}

	// redim inter rows from xs to xd pixels, save in to dest.
	dest.fill(0.0);
	left = xd*xs; a = xd; b = xs;
	s = 0; d = 0; used = 0;
	while (left > 0) {
		used = cimg::min(a,b);
		left -= used; a -= used; b -= used;
		cimg_mapYZV(dest, y, z, v) dest(d, y, z, v) += inter(s, y, z, v)*used;
		if (b == 0) {
			cimg_mapYZV(dest, y, z, v) dest(d, y, z, v) /= xs;
			b = xs; ++d;
		}
		if (a == 0) {
			a = xd;	++s;
		}
	}
	// we are done.
}



/*********************************************************** 
 * getflow is the multiscale "driver" routine
 * it performs the different rescalings, compute the image 
 * differences we need and calls the fixed point solver to 
 * get the flow updates at each level.
 ***********************************************************/
const CImg<float> & PAP::getFlow() 
{
	double ratio = 1.0/scalefactor;
	// we create the flow field and initialize it to 0
	unsigned lx = (unsigned)(nx*pow(ratio, (double)(levels-1))), ly = (unsigned)(ny*pow(ratio, (double)(levels-1))); 
	// we start by initializimg the flow to 0
	U = CImg<float>(lx, ly, nz-1, 2, 0.0);

	// then we run through the pyramid levels
	double xratio = 0.0, yratio = 0.0; // for a more precise flow resize;

	for (int level = levels-1; level >= 0; --level) {

		// we compute the spatial dimension at that level.
		cx = (unsigned)(nx*pow(ratio, level));
		cy = (unsigned)(ny*pow(ratio, level));

		// now we resize the image sequence
		Ir = CImg<float>(cx, cy, nz);
		resize(I,Ir);
		//Ir.display("Resized");

		// and then the flow field
		double mx = (cx*1.0)/(U.width*1.0), my = (cy*1.0)/(U.height*1.0);
#ifndef MATLAB_MEX_FILE       //For Windows use: MEXFILE_FUNCTION
		std::cout << "mx = " << mx << ", my = " << my << std::endl;
#endif
		CImg<float> *Ur = new CImg<float>(cx, cy, nz-1, 2);
		resize(U,*Ur);
		U = *Ur;
		delete Ur;
		//U.resize(cx, cy, nz-1, 2, 3);
		// then apply the interlevel transform to the the flow field: diag(mx, my).*U 
		cimg_mapXYZ(U, x, y, z) {
			U(x, y, z, 0) *= mx;
			U(x, y, z, 1) *= my;
		}

		// now we can compute the tensor entries at that level as well as the fixed flow Jacobian
		computeIUDerivatives();
		// run the fixed point solver for the flow update
		fixedPointSolver();
		// and update. Done at that level
		U += dU;
	}
	return U;
}

/****************************************************
 * compute the warped derivatives as well the fixed 
 * part of the flow jacobian dU
 ****************************************************/
void PAP::computeIUDerivatives()
{
	S = CImg<float>(cx, cy, nz-1, 6);
	cimg_mapXYZ(S, x, y, z) {
		double ux = x + U(x,y,z,0), vy = y + U(x,y,z,1);

		// start by extracting the image values in the spatio-temporal neighborhood of interest
		// use 2 spatial neighborhoods, one centered at (x,y,t), and a similar one at (x+u,y+v,t+1) (x+u =: ux, y+v =: vy)
		double Iuww = Ir.cubic_pix2d(ux-2,vy,z+1), Iuw = Ir.cubic_pix2d(ux-1,vy,z+1);
		double Iuee = Ir.cubic_pix2d(ux+2,vy,z+1), Iue = Ir.cubic_pix2d(ux+1,vy,z+1);
		double Iuss = Ir.cubic_pix2d(ux,vy-2,z+1), Ius = Ir.cubic_pix2d(ux,vy-1,z+1);
		double Iunn = Ir.cubic_pix2d(ux,vy+2,z+1), Iun = Ir.cubic_pix2d(ux,vy+1,z+1);
		double Iunw = Ir.cubic_pix2d(ux-1,vy+1,z+1);
		double Iune = Ir.cubic_pix2d(ux+1,vy+1,z+1);
		double Iuse = Ir.cubic_pix2d(ux+1,vy-1,z+1);
		double Iusw = Ir.cubic_pix2d(ux-1,vy-1,z+1);
		double Iuc  = Ir.cubic_pix2d(ux,vy,z+1);

		double Iww  = Ir.neumann_pix2d(x-2,y,z), Iw = Ir.neumann_pix2d(x-1,y,z);
		double Iee  = Ir.neumann_pix2d(x+2,y,z), Ie = Ir.neumann_pix2d(x+1,y,z);
		double Iss  = Ir.neumann_pix2d(x,y-2,z), Is = Ir.neumann_pix2d(x,y-1,z);
		double Inn  = Ir.neumann_pix2d(x,y+2,z), In = Ir.neumann_pix2d(x,y+1,z);
		double Inw  = Ir.neumann_pix2d(x-1,y+1,z);
		double Ine  = Ir.neumann_pix2d(x+1,y+1,z);
		double Ise  = Ir.neumann_pix2d(x+1,y-1,z);
		double Isw  = Ir.neumann_pix2d(x-1,y-1,z);
		double Ic   = Ir(x,y,z);
		
		// from that define the different image derivatives
		double Iux = (Iuww - 8*Iuw + 8*Iue - Iuee)/12;
		double Ivy = (Iuss - 8*Ius + 8*Iun - Iunn)/12;
		double Ix  = (Iww - 8*Iw + 8*Ie -Iee)/12;
		double Iy  = (Iss - 8*Is + 8*In -Inn)/12;
		double It  = Iuc - Ic;
		double Ixt = Iux - Ix;
		double Iyt = Ivy - Iy;
		double Iuxx = Iuw + Iue - 2*Iuc;
		double Ivyy = Ius + Iun - 2*Iuc;
		double Iuxy = 0.25*(Iune + Iusw - Iunw - Iuse);

		// then the tensor coefficients
		S(x,y,z,0) = sqr(Iux) + gamma*(sqr(Iuxx) + sqr(Iuxy));   // S11
		S(x,y,z,1) = Iux*Ivy  + gamma*Iuxy*(Iuxx + Ivyy);        // S12
		S(x,y,z,2) = Iux*It   + gamma*(Iuxx*Ixt + Iuxy*Iyt);     // S13
		S(x,y,z,3) = sqr(Ivy) + gamma*(sqr(Iuxy) + sqr(Ivyy));   // S22
		S(x,y,z,4) = Ivy*It   + gamma*(Iuxy*Ixt + Ivyy*Iyt);     // S23;
		S(x,y,z,5) = sqr(It)  + gamma*(sqr(Ixt) + sqr(Iyt));     // S33;
	}
	
	//S.display("Tensor");

	JU = CImg<float>(cx, cy, nz-1, 6); // fixed part of the flow derivative. 
	cimg_mapXYZ(JU, x, y, z) { 

		JU(x,y,z,0) = (U.neumann_pix1d(x+1,y,z,0) - U.neumann_pix1d(x-1,y,z,0))/2.0;   // u1x
		JU(x,y,z,1) = (U.neumann_pix2d(x,y+1,z,0) - U.neumann_pix2d(x,y-1,z,0))/2.0;   // u1y
		JU(x,y,z,2) = (U.neumann_pix3d(x,y,z+1,0) - U.neumann_pix3d(x,y,z-1,0))/2.0;   // u1t
		JU(x,y,z,3) = (U.neumann_pix1d(x+1,y,z,1) - U.neumann_pix1d(x-1,y,z,1))/2.0;   // u2x
		JU(x,y,z,4) = (U.neumann_pix2d(x,y+1,z,1) - U.neumann_pix2d(x,y-1,z,1))/2.0;   // u2y
		JU(x,y,z,5) = (U.neumann_pix3d(x,y,z+1,1) - U.neumann_pix3d(x,y,z-1,1))/2.0;   // u2t

	}
}


/*************************************************
 * Compute the coefficients of the linear system *
 * that has to be solved at each fixed point     *
 * iteration.                                    *
 * We need the term \psi'_D(dUk^t S dUk)         *
 * and \psi'_R(||Juk + duk||_F^2)                *
 *************************************************/
void PAP::computeEquationCoefs()
{
	// normally the Data term \psi'_D has been properly dimensioned.
	cimg_mapXYZ(Data, x, y, z) {
		double du1 = dU(x,y,z,0), du2 = dU(x,y,z,1);
		double tmp = sqr(du1)*S(x,y,z,0) + sqr(du2)*S(x,y,z,3) + S(x,y,z,5) + 
					2*(S(x,y,z,1)*du1*du2 + S(x,y,z,2)*du1 + S(x,y,z,4)*du2);
		Data(x,y,z) = 0.5/sqrt(tmp + eps2);
#ifndef MATLAB_MEX_FILE       //For Windows use: MEXFILE_FUNCTION
		checkValue(Data(x,y,z), x, y, z);
#endif
	}
	//Data.display("Data");
	CImg<float> JdU = CImg<float>(cx, cy, nz-1, 6); // variable part of the flow derivative. 
	cimg_mapXYZ(JdU, x, y, z) { 

		JdU(x,y,z,0) = (dU.neumann_pix1d(x+1,y,z,0) - dU.neumann_pix1d(x-1,y,z,0))/2.0;   // du1x
		JdU(x,y,z,1) = (dU.neumann_pix2d(x,y+1,z,0) - dU.neumann_pix2d(x,y-1,z,0))/2.0;   // du1y
		JdU(x,y,z,2) = (dU.neumann_pix3d(x,y,z+1,0) - dU.neumann_pix3d(x,y,z-1,0))/2.0;   // du1t
		JdU(x,y,z,3) = (dU.neumann_pix1d(x+1,y,z,1) - dU.neumann_pix1d(x-1,y,z,1))/2.0;   // du2x
		JdU(x,y,z,4) = (dU.neumann_pix2d(x,y+1,z,1) - dU.neumann_pix2d(x,y-1,z,1))/2.0;   // du2y
		JdU(x,y,z,5) = (dU.neumann_pix3d(x,y,z+1,1) - dU.neumann_pix3d(x,y,z-1,1))/2.0;   // du2t

	}
	
	cimg_mapXYZ(Regul, x, y, z) {
		double tmp = 0;
		// compute the square Frobenius norm of JU + JdU
		for (unsigned v = 0; v < 6; ++v) 
			tmp += sqr(JU(x,y,z,v)) + sqr(JdU(x,y,z,v));
		Regul(x,y,z) = 0.5/sqrt(tmp + eps2);
	}
	//Regul.display("Regul");
}


/***************************************************
 * The fixed point solver compute the flow update  *
 * dU at each resolution iteration.                *
 ***************************************************/
void PAP::fixedPointSolver()
{
	// create the update flow, as a 0 field.
	dU = CImg<float>(cx, cy, nz-1, 2, 0.0);
	
	// create the Data and Regul arrays, with current spatial dimensions
	Data = CImg<float>(cx, cy, nz-1);
	Regul = CImg<float>(cx, cy, nz-1);

	// run a fixed amount of fixed point iterations 
	for (unsigned i = 0; i < fixpoint; ++i) {
		// get the coefs of the linear system to solve at that stage
		computeEquationCoefs();
		// and solve it
		relaxationSolver();
	}
}


/***********************************************
 * the linear solver: Block Gauss-Seidel or    *
 * block SOR.                                  *
 ***********************************************/
void PAP::relaxationSolver() 
{
	unsigned idx = 0; // relaxation loop index, set to 0 to shut the compiler up
	double relError = 0; // relative error between to iterations = mean(||e^{n+1}-e^{n}||_2)

	double omega = 1.9;

#ifndef MATLAB_MEX_FILE       //For Windows use: MEXFILE_FUNCTION
	CImgDisplay disp(U+dU,"flow");
#endif

	for (idx = 0; idx < relax; ++idx) {
		relError = 0.0;
		double dunew = 0, dvnew = 0;
		// loop over all the update flow sites
		cimg_mapXYZ(dU, x, y, z) {

			// extract all the values that will be used to compute the equations
			double A = Data(x,y,z);
			double S11 = S(x,y,z,0);
			double S12 = S(x,y,z,1);
			double S13 = S(x,y,z,2);
			double S22 = S(x,y,z,3);
			double S23 = S(x,y,z,4);
		
			double B  = Regul(x,y,z);
			double BW = Regul.neumann_pix1d(x-1,y,z), BE = Regul.neumann_pix1d(x+1,y,z);
			double BS = Regul.neumann_pix2d(x,y-1,z), BN = Regul.neumann_pix2d(x,y+1,z);
			double BB = Regul.neumann_pix3d(x,y,z-1), BA = Regul.neumann_pix3d(x,y,z+1);

			//double Bw = 0.5*(B + BW), Be = 0.5*(B + BE);
			//double Bs = 0.5*(B + BS), Bn = 0.5*(B + BN);
		    	//double Bb = 0.5*(B + BB), Ba = 0.5*(B + BA);
			double Bw = 2.0/(1.0/B + 1.0/BW), Be = 2.0/(1.0/B + 1.0/BE);
			double Bs = 2.0/(1.0/B + 1.0/BS), Bn = 2.0/(1.0/B + 1.0/BN);
			double Bb = 2.0/(1.0/B + 1.0/BB), Ba = 2.0/(1.0/B + 1.0/BA);

 
			double SB = Bw + Be + Bs + Bn + Bb + Ba;

			double uduw = U.neumann_pix1d(x-1,y,z,0) + dU.neumann_pix1d(x-1,y,z,0), udue = U.neumann_pix1d(x+1,y,z,0) + dU.neumann_pix1d(x+1,y,z,0);
			double udus = U.neumann_pix2d(x,y-1,z,0) + dU.neumann_pix2d(x,y-1,z,0), udun = U.neumann_pix2d(x,y+1,z,0) + dU.neumann_pix2d(x,y+1,z,0);
			double udub = U.neumann_pix3d(x,y,z-1,0) + dU.neumann_pix3d(x,y,z-1,0), udua = U.neumann_pix3d(x,y,z+1,0) + dU.neumann_pix3d(x,y,z+1,0);

			double vduw = U.neumann_pix1d(x-1,y,z,1) + dU.neumann_pix1d(x-1,y,z,1), vdue = U.neumann_pix1d(x+1,y,z,1) + dU.neumann_pix1d(x+1,y,z,1);
			double vdus = U.neumann_pix2d(x,y-1,z,1) + dU.neumann_pix2d(x,y-1,z,1), vdun = U.neumann_pix2d(x,y+1,z,1) + dU.neumann_pix2d(x,y+1,z,1);
			double vdub = U.neumann_pix3d(x,y,z-1,1) + dU.neumann_pix3d(x,y,z-1,1), vdua = U.neumann_pix3d(x,y,z+1,1) + dU.neumann_pix3d(x,y,z+1,1);

			double Uc  = U(x,y,z,0),  Vc  = U(x,y,z,1);
			double dUc = dU(x,y,z,0), dVc = dU(x,y,z,1);

			double a11 = A*S11 + alpha*SB, a12 = A*S12, a22 = A*S22 + alpha*SB;
			double C1 = alpha*(Bw*uduw + Be*udue + Bs*udus + Bn*udun + Bb*udub + Ba*udua -SB*Uc) - A*S13;
			double C2 = alpha*(Bw*vduw + Be*vdue + Bs*vdus + Bn*vdun + Bb*vdub + Ba*vdua -SB*Vc) - A*S23;

			// we have all the system coefficients: solve
			solve2x2(a11, a12, a22, C1, C2, dunew, dvnew);
			dunew = (1.0-omega)*dUc + omega*dunew;
			dvnew = (1.0-omega)*dVc + omega*dvnew;
#ifndef MATLAB_MEX_FILE       //For Windows use: MEXFILE_FUNCTION
			checkValue(dunew, x, y, z);
			checkValue(dvnew, x, y, z);
#endif

			relError += sqr(dUc-dunew) + sqr(dVc-dvnew);
			dU(x,y,z,0) = dunew;
			dU(x,y,z,1) = dvnew;
		}
#ifndef MATLAB_MEX_FILE       //For Windows use: MEXFILE_FUNCTION
		(U+dU).display(disp);
#endif

		// compute the relative error after this pixels run
		relError = sqrt(relError)/(cx*cy*(nz-1.0));
		if (relError < convthres) {  // we have converged :-)
			// can add something clever here
			break;
		}
	}
	// could check wether idx == relax to now wether we have converged
}






/**********************************************************
 * The constructor for the PAP(enberg) class:             *
 **********************************************************/
PAP::PAP(const CImg<float> & I_, 
		 double gamma_, double alpha_, 
		 unsigned levels_, double scalefactor_,
		 unsigned fixpoint_, 
		 unsigned relax_, double convthres_)
		 : I(I_), gamma(gamma_), alpha(alpha_), levels(levels_), scalefactor(scalefactor_), 
		   fixpoint(fixpoint_), relax(relax_), convthres(convthres_), eps2(1e-8)
{
	nx = I.width;
	ny = I.height;
	nz = I.depth;

}