/***************************************************************
 * Header file for the multi fixed point implementation of the *
 * Papenberg et al. optical flow algorithm, formerly known as  *
 * Brox et al.                                                 *
 ***************************************************************/


#ifndef PAP_H_LGJGH876876KJGHKJ7687
#define PAP_H_LGJGH876876KJGHKJ7687

// for the matlab mex function, need the matlab plugin for cimg
#ifdef MATLAB_MEX_FILE       //For Windows use: MEXFILE_FUNCTION
#include <mex.h>
#define cimg_plugin "cimgmatlabplugin.h" 
#endif
#define cimg_lapack 0

// if no matlab function these includes are used for the numerical exceptions
#ifndef MATLAB_MEX_FILE       //For Windows use: MEXFILE_FUNCTION
#include <string>
#include <sstream>
#endif

#include <CImg.h>
using namespace cimg_library;


// exceptions only defined for non matlab version
#ifndef MATLAB_MEX_FILE       //For Windows use: MEXFILE_FUNCTION
class NumericalException {
protected:
	std::string value;
	static std::string toString(unsigned x) 
	{
		std::stringstream ss;
		ss << x;
		std::string res;
		ss >> res;
		return res;
	}
	std::string posString(unsigned x, unsigned y, unsigned z)
	{
		return "(" + toString(x) + "," + toString(y) + "," + toString(z) + ")";
	}
public:
	NumericalException() {}
	NumericalException(unsigned x, unsigned y, unsigned z) 
	{
		value = "Numerical problem at " + posString(x,y,z);
	}
	std::string Value() {return value;}
	virtual ~NumericalException() {}
};



class NaNException : NumericalException {
public:
	NaNException(unsigned x, unsigned y, unsigned z) 
	{
		value = "Got a NaN at " + posString(x, y, z);
	}
};

class INDException : NumericalException {
public:
	INDException(unsigned x, unsigned y, unsigned z)
	{
		value = "Got an indeterminate value at " + posString(x, y, z);
	}
};
#endif

class PAP {

    /* The algorithm variables */

	CImg<float> I;       // this is the input image sequence
	CImg<float> U;       // this will be the computed optical flow

	double gamma;         // gradient constancy constraint weight;
	double alpha;         // smoothness weight;

	unsigned levels;      // number of multiresolution levels;
	double scalefactor;   // approximation of the scaling factor between consecutive levels

	unsigned fixpoint;    // number of fixed point iterations for the increment
	unsigned relax;       // maximum number of relaxation iterations in linear solver
	double convthres;     // threshold on difference of consecutive relaxation result for stopping


	/* variables used in resolution scheme */

	CImg<float> Ir;      // spatially rescaled version of the image sequence I
	CImg<float> JU;      // fixed part (at a given level) for the flow Jacobian
	CImg<float> S;       // the tensor encoding the non linear combination of image derivative

	CImg<float> Data;    // the data coefficient
	CImg<float> Regul;   // the regularization coefficients;
	CImg<float> dU;      // the flow update
	
	unsigned clevel;      // current level in pyramid
	unsigned nx,ny,nz;    // size of the image sequence;
	unsigned cx, cy;      // current spatial size;

	const double eps2;    // normally 0.0001^2
    /* probably needs a few more things... */

public:
	PAP(const CImg<float> & I_, 
		double gamma_, double alpha_, 
		unsigned levels_, double scalefactor_,
		unsigned fixpoint_, 
		unsigned relax_, double convthres_);

	const CImg<float> & getFlow();   // well!

private:
	// because it's usefull to have one:
	static double sqr(double x) {return x*x;}
	// 2x2 symmetric system solver A.X = B
	static void solve2x2(double a11, double a12, double a22, double b1, double b2, double & x, double & y); 
	// use my own sequence spatial resize; 
	static void resize(CImg<float> & source, CImg<float> & dest);

#ifndef MATLAB_MEX_FILE       //For Windows use: MEXFILE_FUNCTION
	// for debuging, test numerical values
	static void checkValue(double val, unsigned x, unsigned y, unsigned z);
#endif

	void computeIUDerivatives();   // Image derivatives and fixed part of the flow at a given level
	void fixedPointSolver();       // the fixed point solver for the flow update
	void computeEquationCoefs();   // data and regularization coefficients at each fixed point iteration
	void relaxationSolver();       // the linear solver at a given fixed point iteration
	

};





#endif // #ifndef PAP_H_LGJGH876876KJGHKJ7687