/***************************************************************
 * The spatio-temporal motion compensated SR class             *
 * Uses a variation on Chanas energy.                          *
 * Implementation file.                                        *
 * Modification of the STDeinterlace for MC DI.                *
 ***************************************************************/


#ifndef MATLAB_MEX_FILE
#include <iostream>
#endif

#include "MCSRgen.h"

/***************************************************
 * SR constructor.                                 *
 * Be sure that all the different image arrays for *
 * the algorithm are properly created with correct *
 * size.                                           *
 ***************************************************/
MCSRgen::MCSRgen(const CImg<float> & HR_, 
				  const CImg<float> & bHRflow_, const CImg<float> & fHRflow_,
				  double lambdas_, double lambdat_, 
				  unsigned fixpoint_, unsigned relax_, double convthres_,
				  unsigned M_, unsigned N_, const CImg<float> & F_)
				  : HR(HR_), bHRflow(bHRflow_), fHRflow(fHRflow_), 
				  lambdas(lambdas_), lambdat(lambdat_),
				  fixpoint(fixpoint_), relax(relax_), convthres(convthres_),
				  M(M_), N(N_), F(F_),
				  eps2(1e-8)
{
	nx = HR.width;
	ny = HR.height;
	nz = HR.depth;
	A = CImg<float>(nx, ny, nz);
	update = Bf = Bb = A;
	initSequence();
}


/*********************************************************************************************
 * initSequence: Initilization of sequencedone in Matlab (HR), but some other init done here *
 *********************************************************************************************/
void MCSRgen::initSequence()
{
	lambda0 = 1.0;

	//setting min and max to allow for minor fluctuation around the uint8 intensity values
	//using [-1, 256] should be followed by a cast back to uint8 [0, 255]
	mi = -1.0;
	Ma = 256.0;
	counter = 0;
}

/********************************************************
 * Computes the coefficients that are used to solve the *
 * linear system at a given fixed point iteration.      *
 ********************************************************/
void MCSRgen::computeCoefficients()
{	
	CImg_3x3(HR, float);
	// a very simple A : no half pixels computations here
	cimg_mapZ(HR, z) cimg_map3x3(HR, x, y, z, 0, HR) {
		A(x,y,z) = 1.0/sqrt(0.25*(sqr(HRnc - HRpc) + sqr(HRcn - HRcp)) + eps2);
	}
	
	// Here, I do not have much choice, do I?
	cimg_mapXYZ(Bf, x, y, z) {

		float HRc = HR(x,y,z);
		float fxn = x + fHRflow.dirichlet_pix3d(x,y,z,0),   fyn = y + fHRflow.dirichlet_pix3d(x,y,z,1);
		Bf(x,y,z) = 1.0/sqrt(sqr(HR.linear_pix3d(fxn,fyn,z+1) - HRc) + eps2);
		float fxp = x + bHRflow.dirichlet_pix3d(x,y,z-1,0), fyp = y + bHRflow.dirichlet_pix3d(x,y,z-1,1);
		Bb(x,y,z) = 1.0/sqrt(sqr(HRc - HR.linear_pix3d(fxp,fyp,z-1)) + eps2);
	}
}


/**********************************************************
 * The following routine is the linear solvers used.      *
 **********************************************************/

void MCSRgen::linearSolver() 
{
	double relError = 0.0;
	double omega = 1.0;   // SOR overrelaxation "parameter"
	
#ifndef MATLAB_MEX_FILE
//	std::cout << "Hello, Inpainting Denoising Solver!" << std::endl;
	//CImgDisplay disp(HR,"scaled");
#endif

	for (unsigned i = 0; i < relax; ++i) {
#ifndef MATLAB_MEX_FILE
	std::cout << "At iteration " << i << std::endl;
#endif
		double mylambdat;
		relError = 0.0;
		float SA, Ac, Aw, An, Ae, As, SB, Bp, Bm, Sw, Sn, Se, Ss, Sm, Sp, Sc, fxn, fyn, fxp, fyp, Snew;//S0, Chic;
		
		// Controlling temporal reg. in first and last frame
		for (int z = 0; z < nz; z++) {  //x++ and ++x different from java, x=0, ++x still gives x=0 inside first loop
			if (z == 0 || z == nz-1) mylambdat = 0;
			else mylambdat = lambdat;

			cimg_mapXY(HR, x, y) {

				// extract coefficients

				// first for the spatial diffusion
				Ac = A(x,y,z);
				//Aw = 0.5*(Ac + A.neumann_pix1d(x-1,y,z)); Ae = 0.5*(Ac + A.neumann_pix1d(x+1,y,z));
				//As = 0.5*(Ac + A.neumann_pix2d(x,y-1,z)); An = 0.5*(Ac + A.neumann_pix2d(x,y+1,z));
				Aw = 2.0/(1.0/Ac + 1.0/A.neumann_pix1d(x-1,y,z)); Ae = 2.0/(1.0/Ac + 1.0/A.neumann_pix1d(x+1,y,z));
				As = 2.0/(1.0/Ac + 1.0/A.neumann_pix2d(x,y-1,z)); An = 2.0/(1.0/Ac + 1.0/A.neumann_pix2d(x,y+1,z));
				SA = Aw + Ae + As + An;

				// then the temporal diffusion
				Bm = Bb(x,y,z); Bp = Bf(x,y,z);
				SB = Bm + Bp;

				// get forward and backward displacements
				fxn = x + fHRflow.dirichlet_pix3d(x,y,z,0);   fyn = y + fHRflow.dirichlet_pix3d(x,y,z,1);
				fxp = x + bHRflow.dirichlet_pix3d(x,y,z-1,0); fyp = y + bHRflow.dirichlet_pix3d(x,y,z-1,1);

				// extract intensity values in the spatial-flow neighborhood
				Sc = HR(x,y,z); 
				Sw = HR.neumann_pix1d(x-1,y,z); Se = HR.neumann_pix1d(x+1,y,z);
				Ss = HR.neumann_pix2d(x,y-1,z); Sn = HR.neumann_pix2d(x,y+1,z);
				Sm = HR.linear_pix3d(fxp, fyp, z-1); Sp = HR.linear_pix3d(fxn, fyn, z+1);

				// compute the relaxation at that pixel     
				Snew = (lambdas*(Aw*Sw + Ae*Se + As*Ss + An*Sn) + mylambdat*(Bm*Sm + Bp*Sp))/(lambdas*SA + mylambdat*SB);
				// Snew becomes the overrelaxation
				//Snew = (1-omega)*Sc + omega*Snew;

				// update relative error and image pixel update value;
				relError += sqr(Snew - Sc);
				update(x,y,z) = HR(x,y,z) - Snew;
				
			}
		}
		//mexPrintf("x%i",M);
		//iterationg through the frames block by block to impose the SR constraint by backprojection
		for (unsigned z = 0; z < nz; z++){
			for (unsigned y = 0; y < ny; y += M){
                for (unsigned x = 0; x < nx; x += N){

					SRconstraint(x,y,z);
				}
			}
		}

		//mexPrintf("number of recursive calls in total: %i\n",counter);
		
#ifndef MATLAB_MEX_FILE
		//HR.display(disp);
#endif
		relError = sqrt(relError)/(nx*ny*nz);
		if (relError < convthres) break;
	}
}
/**********************************************************
 * The following routine imposes the SR constraint by     *
 * backprojection in each 2x2 neighborhood of support.    *
 **********************************************************/

void MCSRgen::SRconstraint(unsigned x, unsigned y, unsigned z)
{
		
	CImg<float> uBlock(N,M); // width, height
	//uBlock.fill(0.0); 
	unsigned curPixVec = 0;
	bool minMaxViolated = false;
	
	//iterating to update each pixel in the block
	for(unsigned bx = 0; bx < N; bx++){				//width
		for(unsigned by = 0; by < M; by++){			//height
			
			unsigned curUpdVec = 0;
			//init
			uBlock(bx,by) = HR(x+bx, y+by, z) - lambda0*update(x+bx, y+by, z);

			// Vectorial coordinate in F of the current pixel
			//curPixVec = M*(bx -1) + by;

			// Iterating to find contributions to the current pixel from the update of the rest
			// of the pixel in the block.
			for(unsigned bbx = 0; bbx < N; bbx++){		//width
				for(unsigned bby = 0; bby < M; bby++){	//height
					
					 
					uBlock(bx,by) += lambda0*update(x+bbx, y+bby, z)*F(curPixVec, curUpdVec);
					curUpdVec++;
				}
			}
			curPixVec++;
			// RECURSIVE STEP! Checking if the values are within bounds (min-max constraint)
			if(uBlock(bx,by)<mi || uBlock(bx,by)>Ma){
				minMaxViolated =  true;
				counter++;
				lambda0 -= 0.1;
				SRconstraint(x,y,z);
				lambda0 = 1.0;
				break;
			}
		}
		if(minMaxViolated == true) break;
	}
	if(minMaxViolated == false){
		//min-max is not violated with the current lambda0, so data is stored in HR
//mexPrintf("in assignment part");
		for(unsigned bx = 0; bx < N; bx++){				//width
			for(unsigned by = 0; by < M; by++){			//height

				HR(x+bx, y+by, z) = uBlock(bx, by);
			}
		}
	} 
}


/****************************************************************
 * Finally the fixed point solver, performing the deinterlacing *
 ****************************************************************/
const CImg<float> & MCSRgen::upscale()
{
	for (unsigned i = 0; i < fixpoint; ++ i) {
		computeCoefficients();
		linearSolver();
		
	}
	return HR;
}