/*****************************************************************
 * cimgmatlabplugin.h
 * --------------
 *
 * cimgmatlabplugin is a "plugin" for the CImg library that allows 
 * to convert CImg<T> images from/to matlab arrays, so that CImg
 * can be used to write matlab mex files.
 * How to use it:
 * file matlabplugin.h must be in a directory that the compiler
 * can locate.
 * Prior to include CImg.h, mex.h must be included first, else it
 * will result in a compiler error.
 * after the inclusion of mex.h, one must define the macro
 * cimg_plugin as "cimgmatlabplugin.h" or <cimgmatlabplugin.h>
 *
 * // my fantastic mex file source code
 * #include <mex.h>
 * ...
 * #define cimg_plugin  "cimgmatlabplugin.h"
 * #include <CImg.h>
 * ...
 *
 *****************************************************************/


#if !defined mex_h
#error the file mex.h must be included!
#endif

#if !defined cimg_version
#error cimgmatlabplugin.h requires that CImg.h is included
#endif




// begin of included methods

private:
    // internally used to transfer matlab matrix (image) values to Image2D
    
    static int isNumericalClassID(mxClassID id) {
	switch (id) {
	case mxLOGICAL_CLASS:
	case mxDOUBLE_CLASS:
	case mxSINGLE_CLASS:
	case mxINT8_CLASS:
	case mxUINT8_CLASS:
	case mxINT16_CLASS:
	case mxUINT16_CLASS:
	case mxINT32_CLASS:
	case mxUINT32_CLASS:
	    return 1;
	default:
	    return 0;
	}
    }



    void makeImageFromMatlabData(const mxArray *matlabArray, mxClassID id) {
	if (id == mxLOGICAL_CLASS) {
	    mxLogical *mdata = mxGetLogicals(matlabArray);
	    cpMatlabData((const mxLogical *)mdata);
	} else {
	    void *mdata = mxGetPr(matlabArray);
	    switch (id) {
	    case mxDOUBLE_CLASS:
		cpMatlabData((const double *)mdata);
		break;
	    case mxSINGLE_CLASS:
		cpMatlabData((const float *)mdata);
		break;
	    case mxINT8_CLASS:
		cpMatlabData((const signed char *)mdata);
		break;
	    case mxUINT8_CLASS:
		cpMatlabData((const unsigned char *)mdata);
		break;
	    case mxINT16_CLASS:
		cpMatlabData((const short *)mdata);
		break;
	    case mxUINT16_CLASS:
		cpMatlabData((const unsigned short *)mdata);
		break;
	    case mxINT32_CLASS:
		cpMatlabData((const int *)mdata);
		break;
	    case mxUINT32_CLASS:
		cpMatlabData((const unsigned int *)mdata);
		break;
	    }
	}
    }
    

    template <typename t> void cpMatlabData(const t* mdata) {
	cimg_mapXYZV(*this, x, y, z, v) {
	    (*this)(x, y, z, v) = (T)(mdata[((v*depth + z)*width+x)*height+y]);
	}
    }

public:

    // get an mxArray
    CImg(const mxArray *mArray, bool vdata = false) {
	int nbdims = mxGetNumberOfDimensions(mArray);
	mxClassID classID = mxGetClassID(mArray);
	if (nbdims > 4 || !isNumericalClassID(classID)) {
	    data=NULL; width=height=depth=dim=0;
#if cimg_debug>1
	    cimg::warn(1,"Matlab array is more than 4D"
                         " or/and not numerical, returning null image.");
#endif	    
	} else {
	    const int *dims = mxGetDimensions(mArray);
	    depth = dim = 1;
	    width = dims[1]; height = dims[0];
	    if (nbdims == 4) {
		depth = dims[2];  
		dim = dims[3];
	    } else if (nbdims == 3) {
		if (vdata) dim = dims[2];
		else depth = dims[2];
	    }
	    data = new T[size()];
	    makeImageFromMatlabData(mArray, classID);
	}
    }



    //! 0perator =. Copy  mxMarray data \p mArray into the current image
    CImg & operator=(const mxArray *mArray) {
	int nbdims = mxGetNumberOfDimensions(mArray);
	int classID = mxGetClassID(matlabArray);
	if (nbdims > 4 || !isNumericalClassID(id)) {
	    delete [] data; data = NULL; width=height=depth=dim=0;
#if cimg_debug>1
	    cimg::warn(1,"Matlab array is more than 4D "
                         "or/and not numerical, returning null image.");
#endif	    
	} else {
	    const int *dims = mxGetDimensions(mArray);
	    depth = dim = 1;
	    width = dims[1]; height = dims[0]; 
	    if (nbdims > 2) depth = dims[2];
	    if (nbdims > 3) dim = dims[3];
	    delete [] data;
	    data = new T[size()];
	    makeImageFromMatlabData(mArray, classID);
	}
    }

private:      
    template <typename c>
    void populate_maltlab_array(c *mdata) const {
	cimg_mapXYZV(*this, x, y, z, v) 
	    mdata[((v*depth + z)*width+x)*height+y] = (c)(*this)(x, y, z, v);
    }
    
    void populate_maltlab_array(mxLogical *mdata) const {
	cimg_mapXYZV(*this, x, y, z, v) 
	    mdata[((v*depth + z)*width+x)*height+y] = (*this)(x, y, z, v)!=0;
    }
    
    
public:
    //! export a CImg image to a matlab array.  
    mxArray *toMatlab(mxClassID id = mxDOUBLE_CLASS, bool squeeze = false) const {
	if (!isNumericalClassID(id)) {
#if cimg_debug>1
	    cimg::warn(1,"Invalid Matlab Class Id Specified.");
#endif
	    return NULL;
	}
	int dims[4];
	dims[0] = height;
	dims[1] = width;
	dims[2] = depth;
	dims[3] = dim;
	if (squeeze && depth == 1) {
	    dims[2] = dim;
	    dims[3] = 1;
	}
	mxArray *mArray = mxCreateNumericArray(4,dims,id,mxREAL);
	if (id == mxLOGICAL_CLASS) {
	    mxLogical *mdata = mxGetLogicals(mArray);
	    populate_maltlab_array(mdata);
	} else {
	    void *mdata = mxGetPr(mArray);
	    switch (id) {
	    case mxDOUBLE_CLASS:
		populate_maltlab_array(( double *)mdata);
		break;
	    case mxSINGLE_CLASS:
		populate_maltlab_array(( float *)mdata);
		break;
	    case mxINT8_CLASS:
		populate_maltlab_array(( signed char *)mdata);
		break;
	    case mxUINT8_CLASS:
		populate_maltlab_array(( unsigned char *)mdata);
		break;
	    case mxINT16_CLASS:
		populate_maltlab_array(( short *)mdata);
		break;
	    case mxUINT16_CLASS:
		populate_maltlab_array(( unsigned short *)mdata);
		break;	
	    case mxINT32_CLASS:
		populate_maltlab_array(( int *)mdata);
		break;
	    case mxUINT32_CLASS:
		populate_maltlab_array(( unsigned int *)mdata);
		break;
	    }	
	}
	return mArray;
    }

// end of included methods
